import { AttributeType, BanditActions, BanditSubjectAttributes, EppoClient, EppoPrecomputedClient, EventDispatcher, Flag, IAssignmentDetails, IAssignmentLogger, IBanditLogger, IContainerExperiment, ObfuscatedFlag } from '@eppo/js-client-sdk-common';
import { IClientConfig, IPrecomputedClientConfig } from './i-client-config';
/**
 * Configuration interface for synchronous client initialization.
 * @public
 */
export interface IClientConfigSync {
    flagsConfiguration: Record<string, Flag | ObfuscatedFlag>;
    assignmentLogger?: IAssignmentLogger;
    banditLogger?: IBanditLogger;
    isObfuscated?: boolean;
    throwOnFailedInitialization?: boolean;
    enableOverrides?: boolean;
    overridesStorageKey?: string;
    configPublishedAt?: string;
    configFetchedAt?: string;
}
export { IClientConfig, IPrecomputedClientConfig };
export { IAssignmentDetails, IAssignmentEvent, IAssignmentLogger, IAsyncStore, Flag, ObfuscatedFlag, EppoAssignmentLogger, IBanditLogger, IBanditEvent, ContextAttributes, BanditSubjectAttributes, BanditActions, Attributes, AttributeType, } from '@eppo/js-client-sdk-common';
export { ChromeStorageEngine } from './chrome-storage-engine';
/**
 * @internal
 */
export declare const NO_OP_EVENT_DISPATCHER: EventDispatcher;
/**
 * Client for assigning experiment variations.
 * @public
 */
export declare class EppoJSClient extends EppoClient {
    /**
     * @deprecated Use `getInstance()` instead.
     */
    static instance: EppoJSClient;
    static initialized: boolean;
    private initialized;
    getStringAssignment(flagKey: string, subjectKey: string, subjectAttributes: Record<string, AttributeType>, defaultValue: string): string;
    getStringAssignmentDetails(flagKey: string, subjectKey: string, subjectAttributes: Record<string, AttributeType>, defaultValue: string): IAssignmentDetails<string>;
    /**
     * @deprecated Use getBooleanAssignment instead
     */
    getBoolAssignment(flagKey: string, subjectKey: string, subjectAttributes: Record<string, AttributeType>, defaultValue: boolean): boolean;
    getBooleanAssignment(flagKey: string, subjectKey: string, subjectAttributes: Record<string, AttributeType>, defaultValue: boolean): boolean;
    getBooleanAssignmentDetails(flagKey: string, subjectKey: string, subjectAttributes: Record<string, AttributeType>, defaultValue: boolean): IAssignmentDetails<boolean>;
    getIntegerAssignment(flagKey: string, subjectKey: string, subjectAttributes: Record<string, AttributeType>, defaultValue: number): number;
    getIntegerAssignmentDetails(flagKey: string, subjectKey: string, subjectAttributes: Record<string, AttributeType>, defaultValue: number): IAssignmentDetails<number>;
    getNumericAssignment(flagKey: string, subjectKey: string, subjectAttributes: Record<string, AttributeType>, defaultValue: number): number;
    getNumericAssignmentDetails(flagKey: string, subjectKey: string, subjectAttributes: Record<string, AttributeType>, defaultValue: number): IAssignmentDetails<number>;
    getJSONAssignment(flagKey: string, subjectKey: string, subjectAttributes: Record<string, AttributeType>, defaultValue: object): object;
    getJSONAssignmentDetails(flagKey: string, subjectKey: string, subjectAttributes: Record<string, AttributeType>, defaultValue: object): IAssignmentDetails<object>;
    getBanditAction(flagKey: string, subjectKey: string, subjectAttributes: BanditSubjectAttributes, actions: BanditActions, defaultValue: string): Omit<IAssignmentDetails<string>, 'evaluationDetails'>;
    getBanditActionDetails(flagKey: string, subjectKey: string, subjectAttributes: BanditSubjectAttributes, actions: BanditActions, defaultValue: string): IAssignmentDetails<string>;
    getExperimentContainerEntry<T>(flagExperiment: IContainerExperiment<T>, subjectKey: string, subjectAttributes: Record<string, AttributeType>): T;
    private ensureInitialized;
    /**
     * @internal
     */
    init(config: Omit<IClientConfig, 'forceReinitialize'>): Promise<EppoJSClient>;
    /**
     * @internal
     */
    offlineInit(config: IClientConfigSync): void;
}
/**
 * Builds a storage key suffix from an API key.
 * @param apiKey - The API key to build the suffix from
 * @returns A string suffix for storage keys
 * @public
 */
export declare function buildStorageKeySuffix(apiKey: string): string;
/**
 * Initializes the Eppo client with configuration parameters.
 *
 * The purpose is for use-cases where the configuration is available from an external process
 * that can bootstrap the SDK.
 *
 * This method should be called once on application startup.
 *
 * @param config - client configuration
 * @returns a singleton client instance
 * @public
 */
export declare function offlineInit(config: IClientConfigSync): EppoClient;
/**
 * Initializes the Eppo client with configuration parameters.
 * This method should be called once on application startup.
 * @param config - client configuration
 * @public
 */
export declare function init(config: IClientConfig): Promise<EppoJSClient>;
/**
 * Used to access a singleton SDK client instance.
 * Use the method after calling init() to initialize the client.
 * @returns a singleton client instance
 * @public
 */
export declare function getInstance(): EppoJSClient;
/**
 * Used to build the URL for fetching the flag configuration.
 * @returns a URL string
 * @public
 */
export declare function getConfigUrl(apiKey: string, baseUrl?: string): string;
/**
 * Client for assigning precomputed experiment variations.
 * @public
 */
export declare class EppoPrecomputedJSClient extends EppoPrecomputedClient {
    static instance: EppoPrecomputedJSClient;
    static initialized: boolean;
    getStringAssignment(flagKey: string, defaultValue: string): string;
    getBooleanAssignment(flagKey: string, defaultValue: boolean): boolean;
    getIntegerAssignment(flagKey: string, defaultValue: number): number;
    getNumericAssignment(flagKey: string, defaultValue: number): number;
    getJSONAssignment(flagKey: string, defaultValue: object): object;
    getBanditAction(flagKey: string, defaultValue: string): Omit<IAssignmentDetails<string>, 'evaluationDetails'>;
    private static getAssignmentInitializationCheck;
}
/**
 * Initializes the Eppo precomputed client with configuration parameters.
 * This method should be called once on application startup.
 * @param config - client configuration
 * @public
 */
export declare function precomputedInit(config: IPrecomputedClientConfig): Promise<EppoPrecomputedClient>;
/**
 * Configuration parameters for initializing the Eppo precomputed client.
 *
 * This interface is used for cases where precomputed assignments are available
 * from an external process that can bootstrap the SDK client.
 *
 * @param precomputedConfiguration - The configuration as a string to bootstrap the client.
 * @param assignmentLogger - Optional logger for assignment events.
 * @param banditLogger - Optional logger for bandit events.
 * @param throwOnFailedInitialization - Optional flag to throw an error if initialization fails.
 * @public
 */
export interface IPrecomputedClientConfigSync {
    precomputedConfiguration: string;
    assignmentLogger?: IAssignmentLogger;
    banditLogger?: IBanditLogger;
    throwOnFailedInitialization?: boolean;
    enableOverrides?: boolean;
    overridesStorageKey?: string;
}
/**
 * Initializes the Eppo precomputed client with configuration parameters.
 *
 * The purpose is for use-cases where the precomputed assignments are available from an external process
 * that can bootstrap the SDK.
 *
 * This method should be called once on application startup.
 *
 * @param config - precomputed client configuration
 * @returns a singleton precomputed client instance
 * @public
 */
export declare function offlinePrecomputedInit(config: IPrecomputedClientConfigSync): EppoPrecomputedClient;
/**
 * Used to access a singleton SDK precomputed client instance.
 * Use the method after calling precomputedInit() to initialize the client.
 * @returns a singleton precomputed client instance
 * @public
 */
export declare function getPrecomputedInstance(): EppoPrecomputedClient;
//# sourceMappingURL=index.d.ts.map